
/**
* The IntegratedModel is a more tightly integrated partial-cache client-server model.
*
* It is a flexible model that can be used for a variety of datasets, including very large
* or rapidly changing datasets. It uses a simple JSON based RPC API for CRUD operations, 
* and optionally a websocket connection for change notifications or full record updates.
*
* It differs from RemoteModel in that it does not fully cache data. Rather, filtering,
* sorting and pagination are conducted server side, allowing the model to only keep the
* records currently utilised by bound controls. To maintain consistency, websockets are 
* used to listen for change notifications. The model can then respond appropriately to
* maintain the desired state.
*
* Interface compatible with RemoteModel although functionally different with loading and 
* fetching.
*
* FIXME: make error
* NOTE 1: This model caches records per fetch, and so cannot be bound to multiple multi-bind
* components without complicating model function with pagination, filtering and more. Attempting 
* to bind to multiple multi-bind components results in a throw error.
*
* NOTE 2: This model does not support in-model field indexes, and such indexes cannot be defined.
* 
*  FIXME: no 'has' support either
* NOTE 3: This model does not support Function based filters or sorts. This model only supports 
* FiltersCollection, FilterType, CriteriaCollection, CriteriaType filter types only, and
* SortCriteria sort types only without Functions.
*/

const IntegratedModel = baseclass => 

/**
 * Mixin to enable a remote model to operate as a partial cache with remote filtering, sorting, pagination
 * @mixin
 * @alias IntegratedModel
 */
class extends baseclass {


  /**
   * Constructor for integrated remote data collection model.
   * @param {SimpleApp}   app                             - The main application object
   * @param {...Object}   confs                           - Configuration objects
   * @param {Boolean}     [confs.versioned=false]         - If record(s) are versioned with the version field
   * @param {String}      [confs.versionField="version"]  - The version field, defaulting to 'version'
   * @param {Object}      confs.url                       - URL configurations
   * @param {String}      [confs.url.websock]             - The URL to use for websocket notifications
   * @param {String}      confs.url.fetch                 - The URL to use for fetch actions
   * @param {Boolean}     [confs.reloadInsert=false]      - Reload fetch upon inserted records
   * @param {Boolean}     [confs.reloadUpdate=false]      - Reload fetch if one fetched record changes
   * @param {Boolean}     [confs.reloadDelete=false]      - Reload fetch if one fetched record is deleted
   * @param {String}      [confs.wsChannel]               - Channel to subscribe to for data notifications
   * @param {Number}      [confs.throttleTime=15000]      - Websocket connection throttle time in ms
   * @param {Number}      [confs.batchSize=0]             - The number of records sent per CUD request. Zero means single batch for all records per operation type.
   */
  constructor( app, ...configs ) {
    const defaults = {
      modelType       : [ 'integrated' ],
      url             : {},
      reloadInsert    : false,
      reloadUpdate    : false,
      reloadDelete    : false,
      wsChannel       : undefined,
      throttleTime    : 15000,
      batchSize       : 0
    };

    const overrides = {
      'autoLoad'      : undefined,
      'indexFields'   : undefined,
    }

    super( app, defaults, ...configs, overrides ); 

    return this;
  }



  /**
   * Setup the model
   */
  setup() {
    super.setup();

    this._ws = undefined;

    this._fetched = false; // Is a fetch current
    this._lastConnect = undefined; // Last timestamp of connection attempt
    this._reconnectTimer = undefined; // Store any connection rate limiter timer
    this._lastQuery = undefined; // Store most recent fetch query
  }



  /**
   * Prepare to subscribe to websockets on a data fetch
   * @async
   * @override
   * @param {Object} [context={}]   - Any context (bind variables) for the dataset's urls
   * @param {Object} [query={}]     - Any supplemental query specific for the subsequent queries
   * @returns {Promise}             - Promise which resolves or throws
   */
  async load( context={}, query={} ) {
    this.app.log( "debug", "Attempting load" );

    if ( this._operation !== undefined )
      this.app.log( "warn", "Load setup performed while another operation in progress" );

    // Store any query and context
    this._query = this.app.util.clone( query );
    this._context = this.app.util.clone( context );

    // Make any required connection
    await this._load();
  }


  /**
   * Attempt connection if utilising websockets
   * @async
   * @protected
   * @throws    - If unable to connect
   */
  async _load() {
    if ( ! this.config.url.websock || ! this.config.wsChannel )
      return;

    this.app.log( "debug", "Connecting to websocket service" );

    try {
      this._operation = "load";
      await this._tryConnect();
      this._operation = undefined;
    }
    catch( err ) {
      this._operation = undefined;
      this.app.log( 'error', `Unable to connect to endpoint: ${err.message}`, err );
      throw new Error( `Unable to connect to endpoint: ${err.message}` );
    }
  }



  /** 
   * Override reload behaviour, connect if not connected
   * @async
   * @override
   */
  async reload() {
    this.app.log( 'debug', `Attempting reload` );

    return this._load();
  }



  /**
   * Fetch one or more records from the remote store and subscribe for changes
   * @async
   * @param {Number} start=0                          - The record number to start from. Defaults to 0;
   * @param {Number} count=100                        - The number of records to fetch. Default is to limit traffic
   * @param {SortsType} [sorts]                       - Any sorts to apply to the fetch
   * @param {FiltersCollection} [filters]             - A collection of filters to apply
   * @returns {Promise<ResultType>}                   - Return records and metadata
   */
  async fetch( start=0, count=100, sorts=[], filters={} ) {
    const retdata = new Array();
    const query = this.app.util.clone( this._query );
    let url, ids, records, response;

    // Build query
    query.operation = "fetch";
    query.filters = {};
    query.sorts = [];
    if ( filters.hasOwnProperty( ':rules' ) && filters[ ':rules' ].length )
      query.filters = this._checkFilters( filters );
    if ( ! Array.isArray( sorts ) )
      sorts = [ sorts ];
    query.sorts = this._checkSorts( sorts );
    query.start = start;
    query.count = count;

    // Store query if needing to re-fetch
    this._lastQuery = query;

    // Attempt the fetch
    try {

      // Clear previously fetched records
      this._clearFetched();

      // Subscribe to change notification websock as able
      await this._subscribe();

      url = this.getUrl( 'fetch' );
      if ( ! url )
        throw new Error( `Fetch URL not defined` );

      // Perform the fetch
      response = await this._remoteRequest( url, 'POST', query );

      // Ensure successful overall response
      if ( ! response || response['success'] != true ) {
        this.app.log( "error", "Fetch failed", response );
        throw new Error( "Fetch failure: " + response.message );
      }

      // Ensure we have records even if empty
      records = response.records;
      if ( ! records || ! response.meta ) {
        this.app.log( "error", "Malformed response", response );
        throw new Error( "Malformed response" );
      }

      // Load into our cache for modifications / removal
      ids = this.parse( records, 'load' );

      for ( let id of ids )
        retdata.push( this.getRecord( id ) );

      // Flag for fetch being present
      this._fetched = true;
      this._loaded = true;
      this._inconsistent = false;

      // Add record metadata
      retdata.filterCount = response.meta.filter;
      retdata.totalCount = response.meta.total;
      retdata.fetchCount = ids.length;
    }
    catch( err ) {
      this._inconsistent = true;
      this._loaded = false;
      this.app.log( "error", "Failure during fetching", err );
      throw new Error( `Failure during fetching: ${err.message}` );
    }

    return retdata;
  }



  /** 
   * Refetch one or more records from remote store with last fetched query
   * @async
   */
  async refetch() {
    const retdata = new Array();
    const query = this.app.util.clone( this._query );
    let url, ids, records, response;

    query = this._lastQuery;
    if ( ! query )
      throw new Error( `Unable to refetch without previous query` );

    // Attempt the fetch
    try {

      // Clear previously fetched records
      this._clearFetched();

      url = this.getUrl( 'fetch' );
      if ( ! url )
        throw new Error( `Fetch URL not defined` );

      // Perform the fetch
      response = await this._remoteRequest( url, 'POST', query );

      // Ensure successful overall response
      if ( ! response || response['success'] != true ) {
        this.app.log( "error", "Refetch failed", response );
        throw new Error( "Refetch failure: " + response.message );
      }

      // Ensure we have records even if empty
      records = response.records;
      if ( ! records || ! response.meta ) {
        this.app.log( "error", "Malformed response", response );
        throw new Error( "Malformed response" );
      }

      // Get all records not in the fetch
      // FIXME incomplete
      

      // Load into our cache for modifications / removal
      ids = this.parse( records, 'load' );

      for ( let id of ids )
        retdata.push( this.getRecord( id ) );

      // Flag for fetch being present
      this._fetched = true;
      this._loaded = true;
      this._inconsistent = false;

      // Add record metadata
      retdata.filterCount = response.meta.filter;
      retdata.totalCount = response.meta.total;
      retdata.fetchCount = ids.length;
    }
    catch( err ) {
      this._inconsistent = true;
      this.app.log( "error", "Failure during fetching", err );
      throw new Error( `Failure during fetching: ${err.message}` );
    }

    return retdata;
  }



  /**
   * Clear this models' records and state
   * @override
   */
  clear( ) {
    this._fetched = false;
    this._data = new Map();
    this._ids = [];
    this._lastConnect = undefined;
    this._modified = [];
    if ( this._reconnectTimer ) {
      clearTimeout(  this._reconnectTimer )
      this._reconnectTimer = undefined;
    }
  }



  /**
   * Clear previously fetched records only
   */
  _clearFetched() {
    const newIds = [];
    const newData = new Map();
    for ( let id of this._ids ) {
      if ( this._data.get( id )._operation == "insert" ) {
        newIds.push( id );
        newData.set( id, this._data.get( id ) );
      }
    }
    this._fetched = false;
    this._data = newData;
    this._ids = newIds;
  }



  /**
   * Handle a notification of a change to one or more records 
   * @async
   * @protected
   * @param {String} changeType         - Type of change, one of insert, update or delete
   * @param {Array<String|Number> ids   - IDs for the change
   */
  async handleNotify( changeType, ids ) {
    let known = [], unknown = [];

    // Find records that are known and unknown
    for ( let testId of ids ) {
      if ( this.hasRecord( testId ) )
        known.push( testId );
      else
        unknown.push( testId );
    }

    // Any update to a record we have results in a reload
    if ( known.length ) {
      return this.reload();
    }

    // Otherwise reload on unknown record changes as directed
    if ( changeType == "insert" ) {
      if ( this.config.reloadInsert && unknown.length )
        return this.reload();
    }
    else if ( changeType == "update" ) {
      if ( this.config.reloadUpdate && unknown.length )
        return this.reload();
    }
    else if ( changeType == "delete" ) {
      if ( this.config.reloadDelete && unknown.length )
        return this.reload();
    }


  }




  /**
   * Manage a websocket connection failure
   * @async
   */
  async handleDisconnect() {
    // Need to reconnect?
    if ( this._fetched ) {
      this.app.log( 'error', `Unexpected connection closure, reconnecting` );
      // Reconnect without raising errors
      try {
        await this._tryConnect();
      }
      catch( err ) {
        this.app.log( 'error', `Reconnection failed: ${err.message}` );
      }
    }
    else {
      this.app.log( 'info', `Unexpected connection closure, reconnection required` );
    }
  }



  /**
   * Subscribe or re-subscribe to websocket channel
   * @async
   * @throws  - If unable to subscribe
   */
  async _subscribe() {
    if ( ! this.config.url.websock || ! this.config.wsChannel )
      return;
    if ( ! this._ws || this._ws.readyState !== 1 )
      throw new Error( "Cannot subscribe while connection not open" );

    await this._trySubscribe();
  }



  /**
   * Attempt to connect to endpoint as needed
   * @async
   * @private
   * @throws    - If unable to connect to websock endpoint
   */
  async _tryConnect() {
    const url = this.getUrl( 'websock' );

    // Already connected?
    if ( this._ws && this._ws.readyState <= 1 ) {
      this.app.log( 'debug', `Endpoint connection already exists` );
      return;
    }

    // Waiting for reconnection?
    if ( this._reconnectTimer )
      throw new Error( `Connection not possible pending throttled reconnection` );

    // Re-connecting too quickly?
    if ( this._lastConnect && ( Date.now() - this._lastConnect ) < this.config.throttleTime ) {
      this.app.log( "warn", `Rapid reconnection, throttling next connection` );
      this._reconnectTimer = setTimeout( () => {
        this._reconnectTimer = undefined;
        this._tryConnect();
      }, 
      this.config.throttleTime );
      throw new Error( `Connection throttled` );
    }

    this.app.log( "debug", "Connecting to websock endpoint", url );

    // Mark connection time for reconnection throttling
    this._lastConnect = Date.now();

    // Get a client handle and connect
    this._ws = this.app.fetchWebsock( url );
    await this._ws.connect();
  }



  /**
   * Attempt to subscribe for change notifications
   * @async
   */
  async _trySubscribe( ) {
    const context = this._context;
    let pieces, channel = this.config.wsChannel;

    // Replace contextual bind variables
    if ( channel.indexOf( ':' ) !== -1 ) {
      pieces = channel.split( '/' );
      pieces.forEach( ( piece, idx ) => {
        if ( piece.charAt( 0 ) == ':' ) {
          let field = piece.replace( /\:/g, '' );
          if ( ! context.hasOwnProperty( field ) ) {
            this.app.log( "error", "Cannot interpolate unknown field", field );
            throw new Error( "Cannot interpolate unknown field: " + field );
          }
          pieces[ idx ] = encodeURIComponent( context[ field ] );
        }
      } );
      channel = pieces.join( '/' );
    }

    this.app.log( "debug", `Subscribing to websock channel ${channel}` );

    await this._ws.subscribe( channel, this );
  }



  /**
   * Check filters are suitable for server side filtering.
   * @param {FiltersCollection|CriteriaCollection} filters  - Filters to check
   * @throws - In event of unsupported filter
   */
  _checkFilters( filters ) {
    const rules = filters[ ':rules' ];
    const mode = filters[ ':mode' ];

    if ( mode && ( mode !== "AND" && mode !== "OR" ) )
      throw new Error( `Unsupported filter mode ${mode}` );

    if ( ! rules )
      return;

    if ( ! Array.isArray( rules ) )
      throw new Error( `Unknown rules type found` );

    for ( let rule of rules ) {

      if ( Array.isArray( rule ) )
        throw new Error( `Unknown filter rule type, expecting object found array` );

      if ( typeof rule == "object" && rule.hasOwnProperty( ':rules' ) ) {
        this._checkFilters( rule );
        continue;
      }

      if ( typeof rule == "object" ) {
        for ( let field in rule ) {

          if ( field == ":rules" ) {
            this._checkFilters( rule[ field ] );
            continue;
          }

          if ( Array.isArray( rule[ field ] ) || typeof rule[ field ] !== "object" ) {
            throw new Error( `Unknown filter field rule type, expecting object` );
          }

          for ( let type in rule[ field ] ) {

            if ( type == "has" )
              throw new Error( `Unsupported value operator 'has'` );

            if ( Array.isArray( rule[ field ][ type ] ) ) {
              for ( let val of rule[ field ][ type ] ) {
                if ( typeof val == "function" )
                  throw new Error( `Unsupported criteria value function` );
              }
            }

            if ( typeof rule[ field ][ type ] == "function" )
              throw new Error( `Unsupported criteria value function` );
          }
        }
      }
    }

    return filters;
  }



  /**
   * Check sorts are suitable for server side sorting
   * @param {SortsType} sorts     - Sorts to check
   */
  _checkSorts( sorts ) {

    for ( let sort of sorts ) {
      if ( typeof sort == "function" )
        throw new Error( `Unsupported sort function` );
      if ( sort.hasOwnProperty( 'func' ) && typeof sort[ 'func' ] == "function" )
        throw new Error( `Unsupported sort function` );
    }

    return sorts;
  }



}

export { IntegratedModel };
