

import { v4 as uuidv4 } from 'uuid';


/**
 * Channel WebSocket Client for the ChannelSock server service
 *
 * Provides a client interface for an aggregated WebSocket connection to a single
 * endpoint making one or more channel subscriptions. 
 */

class ChannelSockClient {


  /**
   * Create a client interface instance
   * @constructor
   * @param {String} endpoint   - Endpoint URL to connect to
   */
  constructor( app, endpoint ) {

    this.app = app;
    this._endpoint = endpoint;
    this._sock = undefined;
    this._models = new WeakMap();
    this._subs = new Map();
    this._subProms = new Set();
    this._unsubProms = new Set();
    this._termProm = new Map();
    this._monitorTimer = undefined;
    return this;
  }


  /**
   * Connect to remote endpoint
   * @async
   * @throws      - In event of connection failure
   */
  async connect() {
    let prom, resolved = false;
    let sock = new WebSocket( this._endpoint );
    this._sock = sock;

    prom = new Promise( ( resolve, reject ) => {

      // Catch errors during connection
      this._sock.addEventListener( 'error', ( err ) => { 
        if ( resolved )
          return;
        return reject( err );
      }, { once: true } )

      this._sock.addEventListener( 'open', () => {
        resolved = true;

        // Connect general handlers
        this._sock.addEventListener( 'message', this._handleMessage.bind( this ) );
        this._sock.addEventListener( 'error', this._handleClose.bind( this, true ) );
        this._sock.addEventListener( 'close' , this._handleClose.bind( this ) );

        return resolve();
      }, { once: true } );

    } );

    return prom;
  }



  /**
   * Proxy connection state method
   * @type {Number}   - Connection state identifier
   */
  get readyState() {
    if ( ! this._sock )
      return 3;
    return this._sock.readyState;
  }
  


  /**
   * Subscribe a model to a channel for changes
   * @async
   * @param {String} channel    - Channel to subscribe to
   * @param {Object} model      - Model that is subscribing
   * @throws                    - If not able to connect or subscribe
   */
  async subscribe( channel, model ) {
    let prom, uuid, prev;
    let chanModelSet = this._subs.get( channel );

    this.app.log( 'debug', `Attempting subscribe to ${channel} by model ${model.uuid}` );

    // Connect as needed
    if ( ! this._sock )
      await this.connect();

    // Setup channel model structure as needed
    if ( ! chanModelSet ) {
      this._subs.set( channel, new Set() );
      chanModelSet = this._subs.get( channel );
    }

    // Model already subscribed? Unsubscribe first
    prev = this._models.get( model );
    if ( prev && prev !== channel ) {
      this.app.log( 'debug', `Attempting unsubscribe before subscribe by model ${model.uuid}` );
      await this.unsubscribe( prev, model );
    }

    // Another model already subscribed to this channel?
    if ( chanModelSet && chanModelSet.size > 0 ) {
      this._models.set( model, channel );
      chanModelSet.add( model );
      return;
    }

    // Add to the set of models for the channel early
    chanModelSet.add( model );
    this._models.set( model, channel );

    // Try the subscribe
    prom = new Promise( ( resolve, reject ) => {
      uuid = uuidv4();

      // Setup a resolver / rejector for the request
      this._subProms.add( { 
        'channel'     : channel,
        'resolver'    : resolve,
        'rejecter'    : reject,
        'model'       : model,
        'id'          : uuid
      } );

      // Attempt the subscribe
      this._sock.send( JSON.stringify( {
        'type'      : "subscribe",
        'channel'   : channel,
        'id'        : uuid
      } ) );
    } );

    return prom;
  }



  /**
   * Unsubscribe a model from a channel
   * @async
   * @param {String} channel  - Channel to unsubscribe from
   * @param {Object} model    - Model to unsubscribe
   */
  async unsubscribe( channel, model ) {
    let chanModelSet = this._subs.get( channel );
    let prom, uuid;

    this.app.log( 'debug', `Attempting unsubscribe from ${channel} by model ${model.uuid}` );

    // Not subscribed to this channel or not connected?
    if ( ! chanModelSet || ! this._sock || this._sock.readyState !== 1 ) {
      if ( chanModelSet ) {
        chanModelSet.delete( model );
        this._models.delete( model );
      }
      if ( ! this._sock || this._sock.readyState !== 1 )
        this._models.delete( model );
      return;
    }

    // Another model still subscribed?
    if ( chanModelSet && chanModelSet.size > 0 ) {
      chanModelSet.delete( model );
      this._models.delete( model );
      return;
    }

    // Try the unsubscribe
    prom = new Promise( ( resolve, reject ) => {

      uuid = uuidv4();

      // Setup a resolver / rejector for the request
      this._unsubProms.add( { 
        'channel'     : channel,
        'resolver'    : resolve,
        'rejecter'    : reject,
        'model'       : model,
        'id'          : uuid
      } );

      // Attempt the unsubscribe
      this._sock.send( JSON.stringify( {
        'type'      : "unsubscribe",
        'channel'   : channel,
        'id'        : uuid
      } ) );
    } );

    return prom;
  }


  /**
   * Terminate the connection and notify models
   * @async
   */
  async terminate() {
    let prom;

    // Already terminated?
    if ( ! this._sock )
      return;

    // Already terminating?
    if ( this._termProm.size )
      return this._termProm.get( 'promise' );

    // Setup promise for completion;
    prom = new Promise( (resolve,reject) => { 
      this._termProm.set( 'resolver', resolve );
      this._termProm.set( 'rejecter', reject );
    } );

    this._termProm.set( 'promise', prom );

    this._sock.terminate();

    return prom;
  }



  /** 
   * Handle an incoming message or reqsponse
   * @param {MessageEvent} ev   - The websocket message event
   */
  _handleMessage( ev ) {
    const msg = JSON.parse( ev.data ) || {};
    let type = msg.type;

    if ( ! type ) {
      this.app.log( "error", "Malformed message", msg );
      throw new Error( "Malformed message" );
    }

    this.app.log( "debug", `${type ? type : "unknown"} message received` );

    switch( type ) {
      case "subscribe":
        return this._handleSubscribe( msg );
      case "unsubscribe":
        return this._handleUnsubscribe( msg );
      case "notify":
        return this._handleNotify( msg );
      case "records":
        return this._handleRecords( msg );
    }
  }


  /**
   * Handle a subscribe reply
   */
  _handleSubscribe( message ) {
    let req;

    // Find the request
    for ( let waiting of this._subProms ) {
      if ( message.id == waiting.id ) {
        req = waiting;
        break;
      }
    }
    if ( ! req ) {
      this.app.log( 'error', `Subscribe reply ${message.id} received without request` );
      return;
    }

    // Subscribe failure
    if ( ! message.success ) {
      this.app.log( 'error', `Subscribe failure, ${message.message}` );
      req.rejecter( message.message );
      this._subProms.delete( req );
      return;
    }

    // Success
    this.app.log( 'debug', `Subscribe to channel ${req.channel} successful` );
    req.resolver();
    this._subProms.delete( req );
  }


  /**
   * Handle record ID change notifications
   * @param {Object} message                    - Incoming update message
   * @param {String} message.channel            - Channel for the change
   * @param {String} message.change             - Change type, of insert, update or delete
   * @param {Array<String|Number>} message.ids  - IDs affected by the change
   */
  _handleNotify( message ) {
    const change = message.change;
    const channel = message.channel;
    const ids = message.ids;
    let chanModelSet;

    if ( channel == undefined || change == undefined )
      throw new Error( `Malformed change notification` );

    chanModelSet = this._subs.get( channel );
    if ( ! chanModelSet ) {
      this.app.log( 'warn', `Change notification for unsubscribed channel` );
      return;
    }

    for ( let model of chanModelSet )
      model.handleNotify( change, ids );
  }


  /**
   * Change complete record changes
   * @param {Object} message                    - Incoming records message
   * @param {String} message.channel            - Channel for the records
   * @param {String} message.records            - Records changed
   */
  _handleRecords( message ) {
    const channel = message.channel;
    const recs = message.records;
    let chanModelSet;

    if ( channel == undefined || recs == undefined || ! recs.length )
      throw new Error( `Malformed record change message` );

    chanModelSet = this._subs.get( channel );
    if ( ! chanModelSet ) {
      this.app.log( "warn", `Records notification for unsubscribed channel` );
      return;
    }

    for ( let model of chanModelSet )
      model.handleRecords( recs );
  }



  /**
   * Handle an unsubscribe reply
   */
  _handleUnsubscribe( message ) {
    let chanModelSet, req;

    // Find the request
    for ( let waiting of this._unsubProms ) {
      if ( message.id == waiting.id ) {
        req = waiting;
        break;
      }
    }
    if ( ! req ) {
      this.app.log( 'error', `Unsubscribe reply ${message.id} received without request` );
      return;
    }

    // Get channel model list
    chanModelSet = this._subs.get( req.channel );
    if ( ! chanModelSet )
      this.app.log( 'warn', `Unsubscribe reply ${message.id} for channel without model set` );

    // Unsubscribe failure is assumed not subscribed, therefore achieving unsubscription
    if ( ! message.success )
      this.app.log( 'warn', `Unsubscribe failure, ${message.message}, assuming not subscribed` );
    else
      this.app.log( 'debug', `Unsubscribe to channel ${req.channel} successful` );

    // Forget the model and resolve
    if ( chanModelSet )
      chanModelSet.delete( req.model );
    this._models.delete( req.model );
    this._unsubProms.delete( req );
    return req.resolver();
  }



  /**
   * Handle connection closure or error (which results in closeure)
   * @param {CloseEvent|Event} ev   - The websocket close or generic event
   * @param {Boolean} errored=false - If the closure if the result of an error
   *
   * NOTE: While this method may be called multiple times in some cases, not all 
   * close events have preceeding error events, and not all error events have 
   * subsequent close events.
   */
  _handleClose( ev, errored=false ) {
    let modelSet, resolver, rejecter;

    // From our own terminate request?
    if ( this._termProm.size ) {
      resolver = this._termProm.get( 'resolver' );
      rejecter = this._termProm.get( 'rejecter' );
      this._termProm.clear();
    }

    // Already errored?
    if ( this._sock == undefined ) {
      if ( resolver )
        resolver();
      return;
    }

    if ( errored )
      this.app.log( 'debug', `Websocket errored and closed: ${ev.reason || ev.message}` );
    else
      this.app.log( 'debug', `Websocket closed ${ev.wasClean ? `cleanly` : `unexpectantly`}: ${ev.reason || ev.message}` );

    // Reset our state
    this._sock = undefined;
    this._models = new WeakMap();
    this._subs = new Map();
    this._subProms = new Set();
    this._unsubProms = new Set();
    this._monitorTimer = undefined;

    // Alert models without waiting
    for ( let channel of this._subs.keys() ) {
      modelSet = this._subs.get( channel );
      for ( let model of modelSet ) {
        model.handleDisconnect()
      }
    }

    if ( errored && rejecter )
      rejecter();
    else if ( resolver )
      resolver();
  }


}

export default ChannelSockClient;
